import json
from io import StringIO
from lura.attrs import ottr
from lura.formats.base import Format

class Encoder(json.JSONEncoder):
  'Custom json encoder which support sets.'

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def default(self, _):
    if isinstance(_, set):
      return list(_)
    return super().default(_)

class Json(Format):
  '''
  Thin wrapper for json. This class configures the json module to use
  ordered dictionaries for backing dicts, and indented json.
  '''

  def __init__(self):
    super().__init__()
    self.object_pairs_hook = ottr

  def loads(self, data, **kwargs):
    'Load json from string `data`.'

    kwargs.setdefault('object_pairs_hook', self.object_pairs_hook)
    return json.loads(data, **kwargs)

  def loadf(self, src, encoding=None, **kwargs):
    'Load json from file `src`.'

    with open(src, encoding=encoding) as fd:
      return self.loadfd(fd, **kwargs)

  def loadfd(self, fd, **kwargs):
    'Load json from file descriptor `fd`.'

    kwargs.setdefault('object_pairs_hook', self.object_pairs_hook)
    return json.load(fd, **kwargs)

  def dumps(self, data, **kwargs):
    'Return dict `data` as json.'

    kwargs.setdefault('cls', Encoder)
    return json.dumps(data, **kwargs)

  def dumpf(self, data, dst, encoding=None, **kwargs):
    'Write dict `data` as json to file `dst`.'

    with open(dst, 'w', encoding=encoding) as fd:
      self.dumpfd(data, fd, **kwargs)

  def dumpfd(self, data, fd, **kwargs):
    'Write dict `data` as json to file descriptor `fd`.'

    kwargs.setdefault('cls', Encoder)
    json.dump(data, fd, **kwargs)
    if hasattr(fd, 'flush') and callable(fd.flush):
      fd.flush()



import jinja2 as jinja2_
import os
from lura.plates.base import Expander

class Jinja2(Expander):
  '''
  Expand templates using Jinja2.

  The following extensions are enabled:

  - jinja2.ext.do
  - jinja2.ext.loopcontrols

  The following environment settings are used:

  - trim_blocks
  - lstrip_blocks
  - FileSystemLoader for cwd
  '''

  extensions = [
    'jinja2.ext.do',
    'jinja2.ext.loopcontrols',
  ]

  def __init__(self, cwd=None):
    super().__init__()
    cwd = os.getcwd() if cwd is None else cwd
    self.engine = jinja2_.Environment(
      trim_blocks=True,
      lstrip_blocks=True,
      loader=jinja2_.FileSystemLoader(cwd),
      extensions=self.extensions,
    )

  def expands(self, env, tmpl):
    '''
    Expand a template.

    :param dict env: expansion environment
    :param str tmpl: template text
    :returns: The expanded template.
    :rtype: str
    '''
    return self.engine.from_string(tmpl).render(env)

  def expandf(self, env, tmpl, dst, encoding=None):
    '''
    Expand a template to a file.

    :param str env: expansion environment
    :param str tmpl: template text
    :param str dst: destination file path
    :returns: True if dst file was written
    :rtype bool:
    '''
    with open(dst, 'w', encoding=encoding) as fd:
      fd.write(self.engine.from_string(tmpl).render(env))

jinja2 = Jinja2()



import os
import stat
from distutils.dir_util import copy_tree
from lura import logs
from pathlib import Path

log = logs.get_logger(__name__)

def isfifo(path):
  return stat.S_ISFIFO(os.stat(path).st_mode)

def mkfifo(path):
  try:
    os.mkfifo(path)
  except FileExistsError:
    if isfifo(path):
      return
    raise

def dump(path, data, mode='w', encoding=None):
  with open(path, mode=mode, encoding=encoding) as fd:
    fd.write(data)
    fd.flush()

def slurp(path, mode='r', encoding=None):
  with open(path, mode=mode, encoding=encoding) as fd:
    return fd.read()

def touch(path, mode=0o600):
  Path(path).touch(mode=mode)

def fext(path):
  if '.' in path:
    return path.rsplit('.', 1)[1]
  raise ValueError(f'File has no extension: {path}')

def flush(file):
  if hasattr(file, 'flush') and callable(file.flush):
    file.flush()

def _tee(source, targets, cond):
  while cond():
    data = source.readline()
    if len(data) == '':
      break
    for target in targets:
      target.write(data)

def tee(source, targets):
  _tee(source, targets, cond=lambda: True)

def Tee(source, targets, name=None):
  from lura.threads import Thread
  class _Tee(Thread):
    def __init__(self):
      super().__init__(target=self.work, name=name)
      self.start()
    def work(self):
      self.work = True
      _tee(source, targets, cond=lambda: self.work is True)
    def stop(self):
      self.work = False
  thread = _Tee()
  return thread

class LineCallbackWriter:

  def __init__(self):
    super().__init__()
    self._log('__init__()')
    self.buf = []

  def _log(self, msg):
    log.noise(f'{type(self).__name__}.{msg}')

  def callback(self, lines):
    self._log('callback()')
    for line in lines:
      print(line)

  def write(self, data):
    self._log(f'write({len(data)})')
    if os.linesep not in data:
      self.buf.append(data)
      return
    line_end, extra = data.rsplit(os.linesep, 1)
    self.buf.append(line_end)
    lines = ''.join(self.buf).splitlines()
    self.buf.clear()
    if extra:
      self.buf.append(extra)
    self.callback(lines)

  def writelines(self, lines):
    self._log(f'writelines({len(lines)})')
    return self.write(''.join(lines))

class LogWriter(LineCallbackWriter):

  def __init__(self, log, level, tag=None):
    super().__init__()
    self.tag = tag
    if isinstance(level, int):
      level = logs.get_level_name(level)
    self.log = getattr(log, level.lower())

  def callback(self, lines):
    self._log(f'callback({len(lines)})')
    for line in lines:
      if self.tag:
        line = f'{self.tag} {line}'
      self.log(line)



import os
import types
from lura.attrs import attr
from collections import MutableMapping, MutableSequence
from copy import deepcopy
from distutils.util import strtobool

def isexc(o):
  '''
  `True` if `o` is a tuple as returned by `sys.exc_info()`, else
  `False`.
  '''

  return isinstance(o, tuple) and len(o) == 3 and (
    isinstance(o[0], type) and
    isinstance(o[1], o[0]) and
    isinstance(o[2], types.TracebackType)
  )

def asbool(val):
  'Turn something (including strings via `strtobool`) into a `bool`.'

  if val == '':
    return False
  return bool(strtobool(val)) if isinstance(val, str) else bool(val)

def remap(src, cls=attr):
  '''
  Recursively convert all MutableMappings found in src to type cls. If
  src is a MutableMapping, then src will also be converted to type cls. This
  is used for mass-converting the mapping type used in a deeply-nested data
  structure, such as converting all dicts to attrs.

  :param [MutableSequence, MutableMapping] src: source collection
  :param type cls: target MutableMapping type
  :returns: a new collection
  :rtype MutableMapping:
  '''
  types = (MutableSequence, MutableMapping)
  if isinstance(src, MutableMapping):
    return cls((
      (k, remap(v, cls)) if isinstance(v, types) else (k, v)
      for (k, v) in src.items()
    ))
  elif isinstance(src, MutableSequence):
    return src.__class__(
      remap(_, cls) if isinstance(_, types) else _
      for _ in src
    )
  else:
    raise ValueError(f'src must be MutableSequence or MutableMapping: {src}')

def merge(a, b):
  'Merge two MutableMappings.'

  assert(isinstance(a, MutableMapping))
  assert(isinstance(b, MutableMapping))
  if not (a or b):
    return type(b)()
  if not a:
    return deepcopy(b)
  if not b:
    return deepcopy(a)
  a = a.copy()
  b = b.copy()
  for k, v in a.items():
    if (
      isinstance(v, MutableMapping) and
      k in b and
      isinstance(b[k], MutableMapping)
    ):
      b[k] = merge(v, b[k])
  a.update(b)
  return a

def scrub(obj, tag='[scrubbed]'):
  'Scrub anything that looks like a password in `MutableMapping` `obj`.'

  from collections.abc import MutableMapping, Sequence
  for name, value in obj.items():
    if isinstance(value, str) and 'pass' in name.lower():
      obj[name] = tag
    elif isinstance(value, bytes) and 'pass' in name.lower():
      obj[name] = tag.encode()
    elif isinstance(value, MutableMapping):
      scrub(value)
    elif isinstance(value, Sequence):
      for item in value:
        if isinstance(value, MutableMapping):
          scrub(value)
  return obj

def common(data, count=None):
  'Return the count most common values found in list data.'

  counts = ((data.count(value), value) for value in set(data))
  common = sorted(counts, reverse=True)
  if count is None:
    return common
  return common[:min(len(data), count)]

class DynamicProxy:
  'Dispatch one method call to a list of object instances.'

  def __init__(self, targets):
    super().__init__()
    self.__targets__ = targets

  def build_dispatch(self, name):
    def dispatch(*args, **kwargs):
      return [
        getattr(target, name)(*args, **kwargs)
        for target in self.__targets__
      ]
    return dispatch

  def __getattr__(self, name):
    return self.build_dispatch(name)

class StrUtil(str):
  'Subclass of string offering extra operations.'

  def lines(self):
    'Strip right newlines and return a split on `os.linesep`.'

    return self.rstrip(os.linesep).splitlines()

  def json(self):
    'Parse as a json object.'

    from lura.formats import json
    return json.loads(self)

  def jsons(self):
    'Parse as a sequence of json objects, one per line.'

    from lura.formats import json
    return [json.loads(blob) for line in self.lines()]

  def yaml(self):
    'Parse as a yaml object.'

    from lura.formats import yaml
    return yaml.loads(self)

  def prefix(self, prefix):
    'Return self with each line prefixed with `prefix`.'

    return os.linesep.join(f'{prefix}{line}' for line in self.split(os.linesep))

  def pipe(self, **kwargs):
    'Spawn a process and write this object to the process stdin.'

    raise NotImplementedError()

strutil = StrUtil



from cryptography.fernet import Fernet

def encrypt(data, key=None):
  key = Fernet.generate_key() if key is None else key
  fernet = Fernet(key)
  ciphertext = fernet.encrypt(data)
  del fernet
  return (ciphertext, key)

def decrypt(data, key):
  fernet = Fernet(key)
  plaintext = fernet.decrypt(data)
  del fernet, key
  return plaintext



import ast
import csv
import os
import sys
from collections.abc import MutableMapping
from io import StringIO
from lura import logs
from lura.attrs import attr, ottr
from lura.formats.base import Format

# FIXME support restkey/restval

log = logs.get_logger(__name__)

class Csv(Format):
  '''
  Serialize or deserialize CSV data with optional type inference on load.

  Dump support is currently not implemented, but will be eventually.

  Load support is implemented by `csv.DictReader`.

  Load-time type inference is implemented by `ast.literal_eval()` and can be
  disabled...

  1. for all fields by passing `infer=False` to the load function.
  2. for a particualr field by passing a type conversion callable for the
     field via `typemap`. (e.g. `lambda _: _` to return the naked value.)
  '''

  def __init__(self):
    super().__init__()

  def loads(
    self, data, infer=True, typemap=None, fieldnames=None, dialect='excel',
    **fmtparams
  ):
    with StringIO(data) as fd:
      return self.loadfd(fd, infer, typemap, fieldnames, dialect, **fmtparams)

  def loadf(
    self, src, encoding=None, infer=True, typemap=None, fieldnames=None,
    dialect='excel', **fmtparams
  ):
    with open(src, 'r', encoding=encoding) as fd:
      return self.loadfd(fd, infer, typemap, fieldnames, dialect, **fmtparams)

  def loadfd(
    self, fd, infer=True, typemap=None, fieldnames=None, dialect='excel',
    **fmtparams
  ):
    msg = 'loadfd({}, infer={}, typemap={}, fieldnames={}, dialect={}'
    log.noise(msg.format(fd, infer, typemap, fieldnames, dialect))
    res = []
    reader = csv.DictReader(fd, fieldnames=fieldnames, dialect=dialect)
    for row in reader:
      for field in row:
        row[field] = self._value(field, row[field], infer, typemap)
      res.append(ottr(row))
    return res

  def dumps(self, data):
    raise NotImplementedError()

  def dumpf(self, data, dst, encoding=None):
    raise NotImplementedError()

  def dumpfd(self, data, fd):
    raise NotImplementedError()

  def _value(self, name, value, infer, typemap):
    if typemap and name in typemap:
      # always use the mapped type if present
      res = typemap[name](value)
      log.noise("_value() typemap: {}: {}('{}') -> {}".format(
        name, typemap[name].__name__, value, res))
      return res
    if not infer:
      # return the naked value if not inferring
      log.noise(f"_value() noinfer: {name}: '{value}'")
      return value
    try:
      # evaluate the value as a literal value
      res = ast.literal_eval(value)
      log.noise("_value() eval: {}: {}('{}') -> {}".format(
        name, res.__class__.__name__, value, res))
      # convert dict to attr for convenience
      if isinstance(res, MutableMapping):
        res = attr(res)
        log.noise(f'_value() {name}: MutableMapping -> attr')
      return res
    except (ValueError, SyntaxError):
      # return the naked value if the value can't be evaluated
      log.noise(f"_value() inferexc: {name}: '{value}'")
      return value



from .json import Json
from .pickle import Pickle
from .yaml import Yaml
from .csv import Csv

json = Json()
pickle = Pickle()
yaml = Yaml()
csv = Csv()

from lura.attrs import attr

ext = attr(
  csv = csv,
  jsn = json,
  json = json,
  pickle = pickle,
  pckl = pickle,
  yaml = yaml,
  yml = yaml,
)

del attr

def loadf(src, encoding=None):
  from lura.io import fext
  format = fext(src).lower()
  if format in formats:
    return formats[format].loadf(src, encoding=encoding)
  msg = "Unsupported extension '.{}' for '{}'; supported: {}"
  raise ValueError(msg.format(format, src, formats))

def dumpf(dst, data, encoding=None):
  from lura.io import fext
  format = fext(dst).lower()
  if format in formats:
    return formats[format].dumpf(dst, data, encoding=encoding)
  msg = "Unsupported extension '.{}' for '{}', supported formats: {}"
  raise ValueError(msg.format(format, src, formats))



import os
import shlex
import sys
import threading
from collections.abc import Sequence
from getpass import getpass
from io import StringIO
from lura import LuraError, formats, logs
from lura.attrs import attr, ottr, wttr
from lura.io import Tee, flush, tee
from lura.shell import shell_path, shjoin, whoami
from lura.sudo import popen as sudo_popen
from lura.utils import StrUtil, scrub
from ptyprocess import PtyProcessUnicode
from subprocess import PIPE, Popen as subp_popen

log = logs.get_logger('lura.run')

def is_non_str_sequence(obj):
  return not isinstance(obj, str) and isinstance(obj, Sequence)

def log_context(log, level=logs.NOISE):
  if not log.isEnabledFor(level):
    return
  scrubbed = scrub(dict(run.context()))
  lines = os.linesep.join(f'    {k}: {v}' for (k, v) in scrubbed.items())
  fn = getattr(log, logs.get_level_name(level).lower())
  fn(lines)

class Result:
  'Returned by run().'

  members = ('args', 'argv', 'code', 'stdout', 'stderr')

  def __init__(self, args, argv, code, stdout, stderr):
    super().__init__()
    self.args = args
    self.argv = argv
    self.code = code
    self.stdout = StrUtil(stdout)
    self.stderr = StrUtil(stderr)

  def as_dict(self, type=ottr):
    return type(((name, getattr(self, name)) for name in self.members))

  def format(self, fmt='yaml'):
    from lura import formats
    tag = 'run.{}'.format(type(self).__name__.lower())
    return formats.ext[fmt].dumps({tag: self.as_dict()})

  def print(self, fmt='yaml', file=None):
    file = sys.stdout if file is None else file
    file.write(self.format(fmt=fmt))
    flush(file)

  def log(self, log, level, fmt='yaml'):
    fn = getattr(log, logs.get_level_name(level).lower())
    fn(self.format(fmt=fmt))

  def lines(self):
    return self.stdout.lines()

  def json(self):
    return self.stdout.json()

  def jsons(self):
    return self.stdout.jsons()

  def yaml(self):
    return self.stdout.yaml()

  def pipe(self, **kwargs):
    return self.stdout.pipe(**kwargs)

class Error(LuraError):
  'Raised by `run()` on unexpected exit code.'

  def __init__(self, result):
    msg = f'Process exited with code {result.code}: {result.args}'
    super().__init__(msg)
    self.result = result

def _run_stdio(proc, args, argv, stdout, stderr):
  log.noise('_run_stdio()')
  out, err = StringIO(), StringIO()
  stdout.append(out)
  stderr.append(err)
  threads = (Tee(proc.stdout, stdout), Tee(proc.stderr, stderr))
  try:
    return run.result(args, argv, proc.wait(), out.getvalue(), err.getvalue())
  finally:
    for thread in threads:
      thread.stop()
      thread.join()
    proc.kill()
    out.close()
    err.close()

def _run_popen(args, argv, env, cwd, shell, stdout, stderr, **kwargs):
  log.noise('_run_popen()')
  proc = subp_popen(
    args if shell else argv, env=env, cwd=cwd, shell=shell, stdout=PIPE,
    stderr=PIPE, encoding='utf-8')
  return _run_stdio(proc, args, argv, stdout, stderr)

def _run_pty(args, argv, env, cwd, shell, stdout, **kwargs):
  log.noise('_run_pty()')
  if shell:
    argv = [run.default_shell, '-c', args]
    args = shjoin(argv)
  proc = PtyProcessUnicode.spawn(argv, env=env, cwd=cwd)
  proc_reader = attr(readline=lambda: f'{proc.readline()[:-2]}{os.linesep}')
  out = StringIO()
  stdout.append(out)
  try:
    try:
      tee(proc_reader, stdout)
    except EOFError:
      pass
    return run.result(args, argv, proc.wait(), out.getvalue(), '')
  finally:
    try:
      proc.kill(9)
    except Exception:
      log.exception('Unhandled exception when finally killing pty process')
    out.close()

def _run_sudo(
  args, argv, env, cwd, shell, stdout, stderr, sudo_user, sudo_group,
  sudo_password, sudo_login, sudo_timeout, **kwargs
):
  log.noise('_run_sudo()')
  proc = sudo_popen(
    args if shell else argv, env=env, cwd=cwd, shell=shell, stdout=PIPE,
    stderr=PIPE, encoding='utf-8', sudo_user=sudo_user, sudo_group=sudo_group,
    sudo_password=sudo_password, sudo_login=sudo_login,
    sudo_timeout=sudo_timeout)
  return _run_stdio(proc, args, argv, stdout, stderr)

def lookup(name):
  log.noise(f'lookup({name})')
  default_value = run.defaults[name]
  context_value = run.context().get(name)
  if is_non_str_sequence(default_value):
    if context_value:
      value = []
      value.extend(default_value)
      value.extend(context_value)
      return value
    else:
      return list(default_value)
  elif context_value is not None:
    return context_value
  else:
    return default_value

def merge_args(user_args):
  log.noise(f'merge_args()')
  stdio = ('stdout', 'stderr')
  for name in run.defaults:
    if name in stdio:
      continue
    if user_args.get(name) is None:
      user_args[name] = lookup(name)
  for name in stdio:
    user_value = user_args.get(name)
    default_value = lookup(name)
    if user_value:
      if not isinstance(user_value, Sequence):
        user_value = (user_value,)
      user_args[name] = []
      user_args[name].extend(user_value)
      user_args[name].extend(default_value)
    else:
      user_args[name] = default_value
  return attr(user_args)

def run(argv, **kwargs):
  log.noise('run() begins with context:')
  log_context(log)
  kwargs = merge_args(kwargs)
  modes = ('popen', 'pty', 'sudo')
  if kwargs.mode not in modes:
    raise ValueError(f"Invalid mode '{kwargs.mode}'. Valid modes: {modes}")
  if isinstance(argv, str):
    args = argv
    argv = shlex.split(args)
  else:
    args = shjoin(argv)
  run_real = globals()[f'_run_{kwargs.mode}']
  result = run_real(args, argv, **kwargs)
  if kwargs.enforce is True and result.code != kwargs.enforce_code:
    raise run.error(result)
  log.noise('run() returns with context:')
  log_context(log)
  return result

def getsudopass(prompt=None):
  log.noise('getsudopass()')
  return getpass(getsudopass.prompt if prompt is None else prompt)

getsudopass.prompt = f'[sudo] password for {whoami()}: '

def run_popen(argv, **kwargs):
  log.noise('run_popen()')
  kwargs['mode'] = 'popen'
  return run(argv, **kwargs)

def run_pty(argv, **kwargs):
  log.noise('run_pty()')
  kwargs['mode'] = 'pty'
  return run(argv, **kwargs)

def run_sudo(argv, **kwargs):
  log.noise('run_sudo()')
  kwargs['mode'] = 'sudo'
  return run(argv, **kwargs)

# modes
run.popen = run_popen
run.pty = run_pty
run.sudo = run_sudo

# results
run.result = Result
run.error = Error

# misc
run.getsudopass = getsudopass
run.default_shell = shell_path()

# defaults
run.defaults = attr()
run.defaults.mode = 'popen'
run.defaults.env = None
run.defaults.cwd = None
run.defaults.shell = None
run.defaults.stdout = []
run.defaults.stderr = []
run.defaults.enforce_code = 0
run.defaults.enforce = True
run.defaults.sudo_user = None
run.defaults.sudo_group = None
run.defaults.sudo_password = None
run.defaults.sudo_login = None
run.defaults.sudo_timeout = 3

# context manager variable storage
run.context = lambda: wttr(run.context.tls.__dict__)
run.context.tls = threading.local()



from .run import run
from .context import Context
from .context import Quash
from .context import Enforce
from .context import Cwd
from .context import Shell
from .context import Env
from .context import Stdio
from .context import Log
from .context import Sudo
from .context import New

run.Context = Context
run.Quash = Quash
run.Enforce = Enforce
run.Cwd = Cwd
run.Shell = Shell
run.Env = Env
run.Stdio = Stdio
run.Log = Log
run.Sudo = Sudo
run.New = New



from .log import Logging

logs = Logging(
  std_logger = __name__,
  std_format = Logging.formats.bare[0],
  std_datefmt = Logging.formats.bare[1],
)

del Logging

class LuraError(RuntimeError):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)



import yaml
from lura.attrs import ottr
from lura.formats import myaml
from lura.formats.base import Format

class DictLoaderMixin:
  'Yaml loader mixin which uses a user-defined dict type.'

  def __init__(self, dict_type):
    object.__init__(self)
    self.dict_type = dict_type

  def construct_yamlmap(self, node):
    data = self.dict_type()
    yield data
    value = self.construct_mapping(node)
    data.update(value)

  def construct_mapping(self, node, deep=False):
    if not isinstance(node, yaml.MappingNode):
      msg = f'Expected MappingNode but found {node.id}'
      raise yaml.constructor.ConstructError(None, None, msg, node.start_mark)
    self.flatten_mapping(node)
    mapping = self.dict_type()
    for key_node, value_node in node.value:
      key = self.construct_object(key_node, deep=deep)
      try:
        hash(key)
      except TypeError as exc:
        raise yaml.constructor.ConstructError(
          'While building mapping',
          node.start_mark,
          f'found unhashable key ({exc})',
          key_node.start_mark,
        )
      value = self.construct_object(value_node, deep=deep)
      mapping[key] = value
    return mapping

class Loader(DictLoaderMixin, yaml.Loader):

  default_dict_type = ottr

  def __init__(self, *args, **kwargs):
    yaml.Loader.__init__(self, *args, **kwargs)
    DictLoaderMixin.__init__(self, self.default_dict_type)
    self.add_constructor(
      'tag:yaml.org,2002:map', type(self).construct_yamlmap
    )
    self.add_constructor(
      'tag:yaml.org,2002:omap', type(self).construct_yamlmap
    )

class SafeLoader(DictLoaderMixin, yaml.SafeLoader):

  default_dict_type = ottr

  def __init__(self, *args, **kwargs):
    yaml.SafeLoader.__init__(self, *args, **kwargs)
    DictLoaderMixin.__init__(self, self.default_dict_type)
    self.add_constructor(
      'tag:yaml.org,2002:map', type(self).construct_yamlmap
    )
    self.add_constructor(
      'tag:yaml.org,2002:omap', type(self).construct_yamlmap
    )

# These dumpers are included for completeness, but are not currently used.
# The myaml module is used for dumping data structures as yaml.

class DictDumperMixin:
  'Yaml dumper mixin which uses a user-defined dict type.'

  def __init__(self):
    object.__init__(self)

  def represent_custom_dict(self, data):
    return self.represent_mapping('tag:yaml.org,2002:map', data.iteritems())

class Dumper(DictDumperMixin, yaml.Dumper):

  default_dict_type = ottr

  def __init__(self, *args, **kwargs):
    yaml.Dumper.__init__(self, *args, **kwargs)
    DictDumperMixin.__init__(self)
    self.add_representer(
      self.default_dict_type, type(self).represent_custom_dict
    )

class SafeDumper(DictDumperMixin, yaml.SafeDumper):

  default_dict_type = ottr

  def __init__(self, *args, dict_type=ottr, **kwargs):
    yaml.SafeDumper.__init__(self, *args, **kwargs)
    DictDumperMixin.__init__(self, dict_type)
    self.add_representer(
      self.default_dict_type, type(self).represent_custom_dict
    )

class Yaml(Format):
  '''
  Thin wrapper for yaml. This class configures the yaml module to use
  ordered dictionaries for backing dicts, prints yaml which is more
  human-readable, and auto-quotes fields containing ``{{`` and ``}}``
  for compatibility with Ansible.
  '''

  # This class uses a modified version of the myaml module to implement
  # the dump*() methods.

  def loads(self, data):
    'Load yaml from string ``data``.'

    return yaml.load(data, Loader=Loader)

  def loadf(self, src, encoding=None):
    'Load yaml from file ``src``.'

    with open(src, encoding=None) as fd:
      return self.loads(fd.read())

  def loadfd(self, fd):
    return self.loads(fd.read()) # FIXME

  def dumps(self, data):
    'Return dict ``data`` as yaml.'

    return myaml.dump(data)

  def dumpf(self, data, dst, encoding=None):
    'Write dict ``data`` as yaml to file ``dst`` using ``encoding``.'

    with open(dst, 'w', encoding=encoding) as fd:
      fd.write(myaml.dump(data))

  def dumpfd(self, data, fd):
    'Write dict ``data`` as yaml to file descriptor ``fd``.'

    fd.write(self.dumps(data))
    if hasattr(fd, 'flush') and callable(fd.flush):
      fd.flush()



from collections.abc import Sequence
from lura import logs
from lura.attrs import attr
from lura.io import LogWriter
from lura.utils import scrub
from .run import run, is_non_str_sequence, log_context

class Context:
  'Base class for run contexts.'

  log = logs.get_logger(__name__)

  @classmethod
  def _log_context(cls):
    log_context(cls.log)

  def __init__(self, autosetvars=None):
    super().__init__()
    self.autosetvars = autosetvars
    self.context = run.context()
    self.previous = attr()
    self.name = type(self).__name__
    self._log(f'__init__(autosetvars={autosetvars})')

  def _log(self, msg, *args, **kwargs):
    self.log.noise(f'{self.name}.{msg}', *args, **kwargs)

  def set(self, name, value, merge=True):
    'Set a context variable and save its previous value for unset().'

    self._log(f'set({name}, merge={merge})')
    assert(name not in self.previous)
    if is_non_str_sequence(value) and merge:
      self.previous[name] = list(self.context.setdefault(name, []))
      self.context[name].extend(value)
    else:
      self.previous[name] = self.context.get(name)
      self.context[name] = value

  def unset(self, name):
    'Restore a context variable to its original value.'

    self._log(f'unset({name})')
    value = self.previous[name]
    if value in (None, []):
      del self.context[name]
    elif is_non_str_sequence(value):
      self.context[name].clear()
      self.context[name].extend(value)
    else:
      self.context[name] = value
    del self.previous[name]

  def autoset(self):
    'Automatically assign context variable values from instance attributes.'

    if not self.autosetvars:
      self._log('autoset() nothing to set')
      return
    self._log('autoset()')
    for name in self.autosetvars:
      merge = True
      if not isinstance(name, str):
        name, merge = name
      self.set(name, getattr(self, name), merge)

  def autounset(self):
    'Restore all previous context variables to their original values.'

    if not self.previous:
      self._log('autounset() nothing to unset')
      return
    self._log('autounset()')
    for name in list(self.previous):
      self.unset(name)

  def cleanup(self):
    'Cleanup any lingering values in run.context.'

    self._log('cleanup()')
    if not all(_ in (None, []) for _ in self.context.values()):
      self.log.debug('run.context is not empty at run.Context cleanup')
      self._log_context()
    self.context.clear()

  def push(self):
    'Increment the thread-global context count.'

    self._log('push()')
    self.context.count = self.context.setdefault('count', 0) + 1

  def pop(self):
    'Decrement the thread-global context count.'

    self._log('pop()')
    self.context.count -= 1
    if self.context.count == 0:
      del self.context['count']
      self.cleanup()

  def __enter__(self):
    self._log('__enter__() begins with context:')
    self._log_context()
    self.push()
    self.autoset()

  def __exit__(self, *exc_info):
    self._log('__exit__()')
    self.autounset()
    self.pop()
    self._log('__exit__() returns with context:')
    self._log_context()

class Quash(Context):

  def __init__(self):
    autosetvars = ('enforce',)
    super().__init__(autosetvars)
    self.enforce = False

class Enforce(Context):

  def __init__(self, enforce_code):
    assert(isinstance(enforce_code, int))
    autosetvars = ('enforce_code', 'enforce')
    super().__init__(autosetvars)
    self.enforce_code = enforce_code
    self.enforce = True

class Cwd(Context):

  def __init__(self, cwd):
    autosetvars = ('cwd',)
    super().__init__(autosetvars)
    self.cwd = cwd

class Shell(Context):

  def __init__(self):
    autosetvars = ('shell',)
    super().__init__(autosetvars)
    self.shell = True

class Env(Context):

  def __init__(self, env):
    autosetvars = ('env',)
    super().__init__(autosetvars)
    self.env = env

class Stdio(Context):

  def __init__(self, stdout, stderr=[], excl=False):
    autosetvars = [('stdout', not excl)]
    if stderr or excl:
      autosetvars.append(('stderr', not excl))
    super().__init__(autosetvars)
    self.stdout = stdout if isinstance(stdout, Sequence) else (stdout,)
    self.stderr = stderr if isinstance(stderr, Sequence) else (stderr,)

class Log(Stdio):

  def __init__(self, log, level='DEBUG', excl=False):
    super().__init__(
      LogWriter(log, level, '[out]'),
      LogWriter(log, level, '[err]'),
      excl,
    )

class Sudo(Context):

  def __init__(
    self, password=None, user=None, group=None, login=None, timeout=None
  ):
    autosetvars = (
      'sudo_user', 'sudo_password', 'sudo_group', 'sudo_login', 'sudo_timeout',
      'mode')
    super().__init__(autosetvars)
    self.sudo_user = user
    self.sudo_password = password
    self.sudo_group = group
    self.sudo_login = login
    self.sudo_timeout = timeout
    self.mode = 'sudo'

class New(Context):

  def __init__(self):
    super().__init__()
    self.old = None

  def __enter__(self):
    super().__enter__()
    filter = ('count',)
    self.old = {k: v for (k, v) in self.context.items() if k not in filter}
    for name in self.old:
      del self.context[name]
    self._log('__enter__() new context:')
    self._log_context()

  def __exit__(self, *exc_info):
    self.context.update(self.old)
    super().__exit__()



from .fstring import FString
from .fstring import fstring
from .jinja2 import Jinja2
from .jinja2 import jinja2
from .strformat import StrFormat
from .strformat import strformat
from .stringtemplate import StringTemplate
from .stringtemplate import stringtemplate

from lura.attrs import attr

engines = attr(
  j2 = jinja2,
  jinja2 = jinja2,

  tmpl = stringtemplate,
  template = stringtemplate,
  stringtemplate = stringtemplate,

  fstr = fstring,
  fstring = fstring,

  fmt = strformat,
  format = strformat,
  strformat = strformat,
)

del attr



def always(value=None):
  'Return a generator that always returns ``value``.'

  while True:
    yield value

forever = always

class BufferedIterator:
  '''
  Iterator wrapper that buffers next() and provides the methods:

  - has_next()
  - peek(default=None)
  '''

  def __init__(self, it):
    super().__init__()
    self._it = iter(it)
    self._buffered = False
    self._next = None

  def __iter__(self):
    return self

  def __next__(self):
    if self._buffered:
      _ = self._next
      self._next = None
      self._buffered = False
      return _
    else:
      return next(self._it)

  def has_next(self):
    'Return False if there are no more items, else True.'

    if self._buffered:
      return True
    try:
      self._next = next(self._it)
      self._buffered = True
    except StopIteration:
      pass
    return self._buffered

  def peek(self, default=None):
    '''
    Return the next item without advancing the iterator, or default if there
    are no more items.
    '''
    return self._next if self.has_next() else default



import sys
import threading
from lura.attrs import attr
from multiprocessing.pool import ThreadPool

def map(thread_count, func, items, chunksize=None):
  with ThreadPool(thread_count) as p:
    return p.map(func, items, chunksize=chunksize)

def imap(thread_count, func, items, chunksize=1):
  with ThreadPool(thread_count) as p:
    return p.imap(func, items, chunksize=chunksize)

pool = attr(
  map = map,
  imap = imap,
)

class Thread(threading.Thread):

  @classmethod
  def spawn(cls, *args, **kwargs):
    thread = cls(*args, **kwargs)
    thread.start()
    return thread

  def __init__(
    self,
    group=None,
    target=None,
    name=None,
    args=(),
    kwargs={},
    *,
    daemon=None,
    reraise=True,
  ):
    super().__init__(group=group, name=name, daemon=daemon, target=self.work__)
    self.target_ = target
    self.args_ = args
    self.kwargs_ = kwargs
    self.result = None
    self.error = None
    self.reraise = reraise

  def work__(self):
    try:
      self.result = self.target_(*self.args_, **self.kwargs_)
    except Exception:
      self.error = sys.exc_info()
      if self.reraise:
        raise



import shlex
import subprocess as subp
from subprocess import PIPE

def sh(argv, encoding='utf-8'):
  proc = subp.Popen(
    argv, stdout=PIPE, stderr=PIPE, shell=True, encoding=encoding)
  try:
    return proc.communicate()
  finally:
    proc.kill()

def shell_path():
  return sh('echo $0')[0].strip()

def whoami():
  return sh('whoami')[0].strip()

def hostname():
  import platform
  return platform.node()

shjoin = subp.list2cmdline



import os
import sys
import subprocess as subp
import threading
from lura import logs
from lura.io import dump, mkfifo, slurp
from lura.shell import shell_path, shjoin
from lura.time import Timer
from tempfile import TemporaryDirectory
from time import sleep

log = logs.get_logger('lura.sudo')
shell = shell_path()
tls = threading.local()

class TimeoutExpired(RuntimeError):

  def __init__(self, sudo):
    self.sudo_argv = shjoin(sudo._sudo_argv())
    self.askpass_argv = sudo._askpass_argv()
    msg = f'Timed out waiting for sudo: {self.sudo_argv}'
    super().__init__(msg)

def _command_argv():
  log.noise('_command_argv()')
  argv = [shjoin(['touch', tls.ok_path]), '&&']
  if not tls.shell:
    argv.append('exec')
  if isinstance(tls.argv, str):
    argv.append(tls.argv)
  else:
    argv.append(shjoin(tls.argv))
  return ' '.join(argv)

def _sudo_argv():
  log.noise('_sudo_argv()')
  sudo_argv = ['sudo', '-A']
  if tls.user is not None:
    sudo_argv += ['-u', tls.user]
  if tls.group is not None:
    sudo_argv += ['-g', tls.group]
  if tls.login:
    sudo_argv.append('-i')
  sudo_argv += [shell, '-c', _command_argv()]
  return sudo_argv

def _askpass_argv():
  log.noise('_askpass_argv()')
  return shjoin([
    sys.executable,
    '-m',
    'lura.sudo', # FIXME
    'askpass',
    tls.fifo_path,
    str(float(tls.timeout)),
  ])

def _check_ok():
 return os.path.isfile(tls.ok_path)

def _make_fifo():
  log.noise('_make_fifo()')
  mkfifo(tls.fifo_path)

def _open_fifo():
  log.noise('_open_fifo()')
  try:
    tls.fifo = os.open(tls.fifo_path, os.O_NONBLOCK | os.O_WRONLY)
    return True
  except OSError:
    return False

def _write_fifo(timeout):
  log.noise('_write_fifo()')
  password = tls.password.encode()
  pos = 0
  end = len(password)
  timer = Timer(start=True)
  while True:
    try:
      n = os.write(tls.fifo, password[pos:])
      log.noise(f'_write_fifo() wrote {n} bytes')
      pos += n
      if pos == end:
        log.noise(f'_write_fifo() write complete')
        return
    except BlockingIOError:
      pass
    if _check_ok():
      log.noise(f'_write_fifo() check ok')
      return
    if tls.timeout < timer.time:
      log.noise(f'_write_fifo() timeout {timeout}s expired')
      raise TimeoutExpired()
    log.noise(f'_write_fifo() sleeps for {tls.sleep_interval}s')
    sleep(tls.sleep_interval)

def _close_fifo():
  log.noise('_close_fifo()')
  try:
    os.close(tls.fifo)
  except Exception:
    log.exception('Error while closing pipe to sudo askpass')
  tls.fifo = None

def _wait_for_sudo():
  log.noise('_wait_for_sudo()')
  timer = Timer(start=True)
  log.noise('_wait_for_sudo() fifo begin')
  while True:
    if _open_fifo():
      try:
        _write_fifo(tls.timeout - timer.time)
        break
      finally:
        _close_fifo()
    if _check_ok():
      log.noise('_wait_for_sudo() check ok 1')
      return
    if tls.timeout < timer.time:
      log.noise(f'_wait_for_sudo() timeout {timeout}s expired 1')
      raise TimeoutExpired()
    log.noise(f'_wait_for_sudo() sleeps for {tls.sleep_interval}s 1')
    sleep(tls.sleep_interval)
  log.noise('_wait_for_sudo() fifo end')
  log.noise('_wait_for_sudo() await ok')
  while not _check_ok():
    if tls.timeout < timer.time:
      log.noise(f'_wait_for_sudo() timeout {timeout}s expired 2')
      raise TimeoutExpired()
    log.noise(f'_wait_for_sudo() sleeps for {tls.sleep_interval}s 2')
    sleep(tls.sleep_interval)
  log.noise('_wait_for_sudo() check ok 2')

def _make_askpass():
  log.noise('_make_askpass()')
  contents = f'#!{shell}\nexec {_askpass_argv()}\n'
  dump(tls.askpass_path, contents)
  os.chmod(tls.askpass_path, 0o700)

def _reset():
  log.noise('_reset()')
  try:
    tls.state_dir_context.__exit__(None, None, None)
  except Exception:
    log.exception('Exception while deleting state directory')
  tls.__dict__.clear()

def _popen():
  log.noise('_popen()')
  _make_fifo()
  _make_askpass()
  tls.env['SUDO_ASKPASS'] = tls.askpass_path
  proc = subp.Popen(
    _sudo_argv(), env=tls.env, cwd=tls.cwd, stdin=tls.stdin,
    stdout=tls.stdout, stderr=tls.stderr, encoding=tls.encoding)
  try:
    _wait_for_sudo()
  except Exception:
    proc.kill()
    raise
  return proc

def popen(
  argv,
  env = None,
  cwd = None,
  shell = None,
  stdin = None,
  stdout = None,
  stderr = None,
  encoding = None,
  text = None,
  sudo_user = None,
  sudo_group = None,
  sudo_login = None,
  sudo_password = None,
  sudo_timeout = None,
):
  try:
    log.noise('popen()')
    tls.state_dir_context = TemporaryDirectory()
    tls.state_dir = tls.state_dir_context.__enter__()
    tls.argv = argv
    tls.env = {} if env is None else env
    tls.cwd = cwd
    tls.shell = False if shell is None else shell
    tls.stdin = stdin
    tls.stdout = stdout
    tls.stderr = stderr
    tls.encoding = encoding
    tls.user = sudo_user
    tls.group = sudo_group
    tls.login = sudo_login
    tls.password = sudo_password
    tls.timeout = 5 if sudo_timeout is None else sudo_timeout
    tls.sleep_interval = 0.1
    tls.askpass_path = os.path.join(tls.state_dir, 'sudo_askpass')
    tls.fifo_path = os.path.join(tls.state_dir, 'sudo_askpass_pipe')
    tls.ok_path = os.path.join(tls.state_dir, 'sudo_ok')
    tls.sudo_ok = False
    tls.fifo = None
    return _popen()
  finally:
    _reset()

Popen = popen



from abc import abstractmethod

class BoundExpander:
  'Bind an expander to an environment.'

  def __init__(self, expander, env):
    super().__init__()
    self.expander = expander
    self.env = env

  def expands(self, tmpl):
    return self.expander.expands(self.env, tmpl)

  def expandf(self, tmpl, dst):
    return self.expander.expandf(self.env, tmpl, dst)

  def expandfs(self, src):
    return self.expander.expandfs(self.env, src)

  def expandff(self, src, dst):
    return self.expander.expandff(self.env, src, dst)

class Expander:

  def __init__(self, *args, **kwargs):
    super().__init__()

  def bind(self, env):
    return BoundExpander(self, env)

  @abstractmethod
  def expands(self, env, tmpl):
    '''
    Expand a template.

    :param dict env: expansion environment
    :param str tmpl: template text
    :returns: The expanded template.
    :rtype: str
    '''
    pass

  @abstractmethod
  def expandf(self, env, tmpl, dst, encoding=None):
    '''
    Expand a template to a file.

    :param str env: expansion environment
    :param str tmpl: template text
    :param str dst: destination file path
    :returns: True if dst file was written
    :rtype bool:
    '''
    pass

  def expandfs(self, env, src, encoding=None):
    '''
    Expand a template file to a string.

    :param dict env: expansion environment
    :param str src: template file path
    :returns: The expanded template.
    :rtype: str
    '''
    with open(src, 'r', encoding=encoding) as fd:
      tmpl = fd.read()
    return self.expands(env, tmpl)

  def expandff(self, env, src, dst, encoding=None):
    '''
    Expand a template file to another file.

    :param str env: expansion environment
    :param str src: template source file path
    :param str dst: destination file path
    :returns: True if dst file was written
    :rtype bool:
    '''
    with open(src, 'r', encoding=encoding) as fd:
      tmpl = fd.read()
    self.expandf(env, tmpl, dst, encoding=encoding)



import click
import os
import sys
import threading
import traceback
from lura import sudo
from lura.crypto import decrypt
from lura.io import mkfifo, slurp, touch
from lura.utils import asbool

@click.group()
def cli():
  pass

@cli.command()
@click.option('-u', '--user', help='Target user.')
@click.option('-g', '--group', help='Target group.')
@click.option('-i', '--login', is_flag=True, help='Run as login shell.')
@click.argument('argv', nargs=-1)
def run(user, group, login, argv):
  timeout = float(os.environ.get('LURA_SUDO_TIMEOUT', '5.0'))
  file = os.environ['LURA_SUDO_FILE']
  key = os.environ['LURA_SUDO_KEY']
  ok = os.environ.get('LURA_SUDO_OK')
  keep = asbool(os.environ.get('LURA_SUDO_KEEP', '0'))
  for _ in (
    'LURA_SUDO_FILE', 'LURA_SUDO_KEY', 'LURA_SUDO_OK', 'LURA_SUDO_KEEP',
    'LURA_SUDO_TIMEOUT',
  ):
    if _ in os.environ:
      del os.environ[_]
  password = decrypt(slurp(file).encode(), key.encode()).decode()
  key = None
  if not keep:
    os.unlink(file)
  file = None
  process = sudo.popen(
    argv,
    stdin = sys.stdin,
    stdout = sys.stdout,
    stderr = sys.stderr,
    become_login = login,
    become_user = user,
    become_group = group,
    become_password = password,
    become_timeout = timeout,
  )
  password = None
  try:
    if ok:
      touch(ok)
    code = process.wait()
    sys.exit(code)
  finally:
    process.kill()

@cli.command()
@click.argument('fifo')
@click.argument('timeout', type=float)
def askpass(fifo, timeout):
  def on_timeout():
    try:
      raise RuntimeError(f'Timed out reading become password from fifo: {fifo}')
    except RuntimeError:
      traceback.print_exc()
    os._exit(1)
  mkfifo(fifo)
  threading.Timer(timeout, on_timeout).start()
  password = slurp(fifo)
  sys.stdout.write(password)
  sys.stdout.flush()
  os._exit(0)

if __name__ == '__main__':
  from lura.sudo.__main__ import cli
  cli()



import sys
import traceback
from lura import logs

class Hooks:

  log = logs.get_logger(__name__)

  def __init__(
    self,
    hooks = None,
  ):
    super().__init__()
    self.log.noise('__init__()')
    self.hooks = []
    if hooks:
      self.hooks.extend(hooks)

  def __iter__(self):
    self.log.noise('__iter__()')
    return self.hooks.__iter__()

  def handler(self, source, hook, signal, ev):
    self.log.noise(f'handler({source}, {signal}, {ev})')
    return getattr(hook, signal, None)

  def format_error(self, source, hook, signal, ev):
    self.log.noise('format_error()')
    msg = msg = "Hook '{}' raised exception for signal '{}', event '{}'"
    return msg.format(hook, signal, ev)

  def error(self, source, hook, signal, ev):
    self.log.noise('error()')
    print(self.format_error(source, hook, signal, ev), file=sys.stderr)
    traceback.print_exc()

  def missing(self, source, hook, signal, ev):
    self.log.noise(f'missing({source}, {hook}, {signal}, {ev})')
    # noop

  def __getattr__(self, signal):
    self.log.noise(f'__getattr__({signal})')
    def dispatch(source, ev, *args, **kwargs):
      self.log.noise(f'dispatch({source}, {ev})')
      for hook in self.hooks:
        fn = self.handler(source, hook, signal, ev)
        if not fn:
          self.missing(source, hook, signal, ev)
          continue
        try:
          fn(source, ev, *args, **kwargs)
        except Exception:
          self.error(source, hook, signal, ev)
    return dispatch

  def add(self, hook):
    self.log.noise(f'add({hook})')
    if hook not in self.hooks:
      self.hooks.append(hook)

  def remove(self, hook):
    self.log.noise(f'remove({hook})')
    if hook in self.hooks:
      self.hooks.remove(hook)

  def clear(self):
    self.log.noise('clear()')
    self.hooks.clear()



from lura.plates.base import Expander

class StrFormat(Expander):
  'Expand templates using str.format().'

  def __init__(self):
    super().__init__()

  def expands(self, env, tmpl):
    '''
    Expand a template.

    :param dict env: expansion environment
    :param str tmpl: template text
    :returns: The expanded template.
    :rtype: str
    '''
    return tmpl.format(**env)

  def expandf(self, env, tmpl, dst, encoding=None):
    '''
    Expand a template to a file.

    :param str env: expansion environment
    :param str tmpl: template text
    :param str dst: destination file path
    :returns: True if dst file was written
    :rtype bool:
    '''
    with open(dst, 'w', encoding=encoding) as fd:
      fd.write(tmpl.format(**env))

strformat = StrFormat()



import random, math

def rrange(start, stop=None, step=None):
  '''
  Return a randomized range of integers using a linear congruential
  pseudo-random number generator.
  '''
  # Couresty of Thomas Lux: http://bit.ly/2FcrJKY

  # Set a default values the same way "range" does.
  if (stop == None):
    start, stop = 0, start
  if (step == None):
    step = 1
  # Use a mapping to convert a standard range into the desired range.
  mapping = lambda i: (i*step) + start
  # Compute the number of numbers in this range.
  maximum = (stop - start) // step
  # Seed range with a random integer.
  value = random.randint(0, maximum)
  # Construct an offset, multiplier, and modulus for a linear
  # congruential generator. These generators are cyclic and
  # non-repeating when they maintain the properties:
  #
  #   1) "modulus" and "offset" are relatively prime.
  #   2) ["multiplier" - 1] is divisible by all prime factors of "modulus".
  #   3) ["multiplier" - 1] is divisible by 4 if "modulus" is divisible by 4.
  #
  # Pick a random odd-valued offset.
  offset = random.randint(0,maximum) * 2 + 1
  # Pick a multiplier 1 greater than a multiple of 4.
  multiplier = 4*(maximum//4) + 1
  # Pick a modulus just big enough to generate all numbers (power of 2).
  modulus = int(2**math.ceil(math.log2(maximum)))
  # Track how many random numbers have been returned.
  found = 0
  while found < maximum:
    # If this is a valid value, yield it in generator fashion.
    if value < maximum:
      found += 1
      yield mapping(value)
    # Calculate the next value in the sequence.
    value = (value*multiplier + offset) % modulus



import os
import requests
import statistics as stats
import sys
import time
import traceback
from requests.exceptions import ConnectTimeout, ReadTimeout, Timeout
from collections import OrderedDict
from collections.abc import Mapping, Sequence
from lura import logs
from lura.formats import yaml
from lura.attrs import attr, ottr
from lura.threads import pool
from lura.utils import common

log = logs.get_logger(__name__)

class Base(ottr):

  def __init__(self):
    super().__init__()

  def format(self, prefix=''):
    name = type(self).__name__.lower()
    text = yaml.dumps({name: self})
    if not prefix:
      return text
    res = os.linesep.join(
      f'{prefix}{line}' for line in text.rstrip(os.linesep).splitlines()
    )
    print(repr(res))
    return res

  def print(self, prefix='', *args, **kwargs):
    print(self.format(prefix), *args, **kwargs)

class Request(Base):

  def __init__(self, endpoint, type, headers, data):
    super().__init__()
    self.endpoint = endpoint
    self.type = type
    self.headers = headers
    self.data = data

class Response(Base):

  def __init__(self, headers, code, text, exc_info, start, end):
    super().__init__()
    self.headers = headers
    self.code = code
    self.text = text
    self.exc_info = exc_info
    if exc_info is not None:
      exc_info = ottr(
        type = '{}.{}'.format(exc_info[0].__module__, exc_info[0].__name__),
        value = repr(exc_info[1]),
        traceback = ''.join(traceback.format_exception(*exc_info)),
      )
    self.exc_info = exc_info
    self.start = start
    self.end = end

class Result(Base):

  def __init__(self, id, request, response, start, end):
    super().__init__()
    self.id = id
    self.request = request
    self.response = response
    self.start = start
    self.end = end

class ResultSet(Base):

  def __init__(self, results, start, end):
    super().__init__()
    self.results = results
    self.start = start
    self.end = end

class ConCurl:

  def __init__(
    self,
    endpoint,
    request_type,
    request_headers = None,
    data = None,
    thread_count = None,
    request_count = None,
    response_timeout = None,
    print_dots = None,
  ):
    super().__init__()
    self.endpoint = endpoint
    self.request_type = request_type
    if request_headers is not None:
      request_headers = self.parse_request_headers(request_headers)
    self.request_headers = request_headers
    self.data = data
    self.thread_count = 1 if thread_count is None else thread_count
    self.request_count = 10 if request_count is None else request_count
    self.response_timeout = 5 if response_timeout is None else response_timeout
    self.print_dots = True if print_dots is None else print_dots
    self.column = 0

  def parse_request_headers(self, request_headers):
    if isinstance(request_headers, Mapping):
      return request_headers
    elif isinstance(request_headers, Sequence):
      return dict((header.split(': ', 1) for header in request_headers))
    else:
      raise ValueError(f'Type not supported: {request_headers}')

  def build_request_endpoint(self, *args, **kwargs):
    return self.endpoint

  def build_request_type(self, *args, **kwargs):
    return self.request_type

  def build_request_headers(self, *args, **kwargs):
    return self.request_headers

  def build_request_data(self, *args, **kwargs):
    return self.data

  def build_request(self, id):
    endpoint = self.build_request_endpoint(id=id)
    type = self.build_request_type(id=id, endpoint=endpoint)
    headers = self.build_request_headers(id=id, endpoint=endpoint, type=type)
    data = self.build_request_data(
      id=id, endpoint=endpoint, type=type, headers=headers)
    return self.Request(endpoint, type, headers, data)

  def build_requests(self):
    return ((id, self.build_request(id)) for id in range(self.request_count))

  def build_request_call_args(self, request):
    args = (request.endpoint,)
    kwargs = dict(
      headers = request.headers,
      timeout = self.response_timeout,
      data = request.data
    )
    return args, kwargs

  def print_dot(self, code, exc):
    if not self.print_dots:
      return
    if code == 200 and exc is None:
      c = '.'
    elif code != 200 and exc is None:
      c = '!'
    elif isinstance(exc, ConnectTimeout):
      c = 'c'
    elif isinstance(exc, ReadTimeout):
      c = 'r'
    elif isinstance(exc, Timeout):
      c = 't'
    elif exc:
      c = 'x'
    else:
      c = '?'
    endl = ''
    self.column += 1
    if self.column == 80:
      endl = os.linesep
      self.column = 0
    print(c, end=endl, flush=True)

  def request(self, request):
    headers = code = text = exc_info = exc = None
    start = time.time()
    end = None
    try:
      action = request.type.lower()
      call = getattr(requests, action)
      args, kwargs = self.build_request_call_args(request)
      log.noise(f'request() {action}(*{args} **{kwargs})')
      response = call(*args, **kwargs)
      end = time.time()
      headers = dict(response.headers)
      code = response.status_code
      text = response.text
    except Exception as _:
      end = time.time() if end is None else end
      exc_info = sys.exc_info()
      exc = _
    self.print_dot(code, exc)
    return self.Response(headers, code, text, exc_info, start, end)

  def test(self, args):
    start = time.time()
    id, request = args
    response = self.request(request)
    return self.Result(id, request, response, start, time.time())

  def handle_results(self, results):
    start, end, results = results.start, results.end, results.results
    args = attr()
    args.start = start
    args.end = end
    args.runtime = end - start
    args.res = results
    args.res_200 = [
      res for res in results
      if res.response.code == 200 and res.response.exc_info is None
    ]
    args.res_not_200 = [
      res for res in results
      if res.response.code != 200 and res.response.exc_info is None
    ]
    args.res_timeouts = [
      res for res in results
      if res.response.exc_info is not None and
        res.response.exc_info.type.endswith('Timeout') # FIXME
    ]
    args.res_exc = [
      res for res in results
      if res.response.exc_info is not None and
        not res.response.exc_info.type.endswith('Timeout') # FIXME
    ]
    args.res_bad = args.res_not_200 + args.res_exc
    args.res_num = len(args.res)
    args.res_num_succeeded = len(args.res_200)
    args.res_num_failed = len(args.res_not_200)
    args.res_num_timeouts = len(args.res_timeouts)
    args.res_num_raised = len(args.res_exc)
    args.res_times = [
      res.response.end - res.response.start for res in args.res_200
    ]
    times = args.times = attr()
    names = ('mean', 'harmonic_mean', 'median')
    if args.res_times:
      times.mean = stats.mean(args.res_times)
      times.median = stats.median(args.res_times)
      times.frequency = common([int(_) for _ in args.res_times])
      times.common = times.frequency[:min(len(times.frequency), 6)]
      times.uncommon = times.frequency[-min(len(times.frequency), 6):]
    else:
      times.mean = 0.0
      times.median = 0.0
      times.frequency = [(0, 0)]
      times.common = [(0, 0)]
      times.uncommon = [(0, 0)]
    return args

  def run(self):
    start = time.time()
    if self.print_dots:
      self.column = 0
      print(
        '.=200 !=Not 200 c/r/t=Connect/Read/Other timeout x=Exception ' + \
        '?=Unknown error')
    thread_count = min(self.thread_count, self.request_count)
    results = pool.map(thread_count, self.test, self.build_requests())
    end = time.time()
    if self.print_dots:
      print(flush=True)
    return self.handle_results(ResultSet(results, start, end))

ConCurl.Request = Request
ConCurl.Response = Response
ConCurl.Result = Result



import os
import sys
from abc import abstractmethod
from lura.hash import hashs
from lura.utils import merge

class Format:
  'Serialize or deserialize data.'

  def __init__(self, *args, **kwargs):
    super().__init__()

  @abstractmethod
  def loads(self, data):
    pass

  @abstractmethod
  def loadf(self, src, encoding=None):
    pass

  @abstractmethod
  def loadfd(self, fd):
    pass

  @abstractmethod
  def dumps(self, data):
    pass

  @abstractmethod
  def dumpf(self, data, dst, encoding=None):
    pass

  @abstractmethod
  def dumpfd(self, data, fd):
    pass

  def mergef(self, path, patch, encoding=None):
    '''
    Merge data into an existing file.

    :param dict patch: data to merge
    :param str path: path to file containing data to load and merge with patch

    1. Read ``path`` file contents into data dict
    2. Merge ``patch`` into data dict
    3. Write write merged data dict back to ``path``, if needed

    If ``path`` doesn't exist, then ``patch`` is dumped to ``path``.
    '''
    if not os.path.isfile(path):
      return self.dumpf(patch, path, encoding=encoding)
    data = self.loadf(path)
    data_hash = hashs(self.dumps(data))
    merged = merge(data, patch)
    merged_str = self.dumps(merged)
    merged_hash = hashs(merged_str)
    if data_hash == merged_hash:
      return False
    with open(path, 'w', encoding=encoding) as fd:
      write(merged_str)

  def mergeff(self, path, patch, encoding=None):
    'Merge data from file patch into data at file path.'

    return self.mergef(self.loads(patch), path, encoding=encoding)

  def print(self, data, *args, **kwargs):
    print(self.dumps(data).rstrip(), *args, **kwargs)



import time
from lura.iter import BufferedIterator, forever

class Timer:

  def __init__(self, start=False):
    super().__init__()
    self.begin = None
    self.end = None
    if start:
      self.start()

  def __enter__(self):
    self.start()
    return self

  def __exit__(self, *exc_info):
    self.stop()

  def start(self):
    self.begin = time.time()
    self.end = None

  def stop(self):
    end = time.time()
    if self.begin is None:
      raise Value('Timer not started')
    self.end = end

  @property
  def started(self):
    return self.start is not None and self.end is None

  @property
  def time(self):
    now = time.time()
    if self.begin is None:
      raise ValueError('Timer not started')
    return (now if self.end is None else self.end) - self.begin

  def format(self, precision=10):
    return f'%0.{precision}f' % self.time

  def print(self, **kwargs):
    print(self.format(), **kwargs)

def poll(test, timeout=-1, retries=-1, pause=0.0):
  '''
  Poll for a condition.

  :param callable test: test returning True for pass and False for fail
  :param float timeout: -1 or return False after this many seconds
  :param int retries: -1 or return False after this many retries
  :param float pause: -1 or number of seconds to wait between retries
  :returns: True if the test succeeded else False
  :rtype bool:
  '''
  timer = Timer(start=True) if timeout >= 0 else None
  tries = BufferedIterator(forever() if retries < 0 else range(-1, retries))
  for _ in tries:
    if test():
      return True
    if timer and timer.time >= timeout:
      break
    if pause > 0 and tries.has_next():
      time.sleep(pause)
  return False



import hashlib

def hashs(buf, alg='sha512'):
  'Hash a buffer using hashlib.'

  algs = hashlib.algorithms_guaranteed
  if not alg in algs:
    raise ValueError(f'Algorithm {alg} not in {algs}')
  o = getattr(hashlib, alg)()
  if isinstance(buf, str):
    buf = buf.encode("ascii")
  o.update(buf)
  return str(o.hexdigest())

def hashf(path, alg='sha512'):
  'Hash a file using hashlib.'

  algs = hashlib.algorithms_guaranteed
  if not alg in algs:
    raise ValueError(f'Algorithm {alg} not in {algs}')
  o = getattr(hashlib, alg)()
  with open(path, 'rb') as fd:
    while True:
      buf = fd.read(hashf.buflen)
      if len(buf) == 0:
        break
      o.update(buf)
  return str(o.hexdigest())

hashf.buflen = 256 * 1024

def checkhash(path, alg, hash):
  '''
  Calculate the hash for a file at ``path`` using algorithm ``alg`` and
  compare it to a known hash ``sum``. If the hashes do not match, then
  ``checkhash.error`` is raised.
  '''

  hash2 = hashf(path, alg)
  if hash != hash2:
    raise checkhash.error(path, alg, hash, hash2)

class HashError(Exception):

  def __init__(self, path, alg, expected, received):
    msg = f'{path}: expected {alg} {expected}, got {received}'
    super().__init__(msg)
    self.path = path
    self.alg = alg
    self.expected = expected
    self.received = received

checkhash.error = HashError



'''
`module layout`
```
automation/
  __init__.py
  __assets__/
    installer/
      nginx.conf.j2
  installer.py
```

`installer.py`
```
from lura import assets
from lura.run import run
from lura.plates import jinja2

class Install:

  def install(self):
    run('apt-get install -y nginx')
    conf_tmpl_path = assets.path(self, 'nginx.conf.j2')
    env = {'server_name': 'www.funhub.com'}
    jinja2.expandff(env, conf_tmpl_path, '/etc/nginx/sites-enabled/funhub.com')
    run('systemctl reload nginx')
```
'''

import os
import inspect
import types
from importlib import import_module
from lura import logs, plates

log = logs.get_logger(__name__)

def resolve(module):
  'Return the absolute path to the assets directory for a module.'

  if not hasattr(module, '__file__'):
    log.noise(f'resolve({module}) __file__ missing')
    return None
  mod_dir = os.path.dirname(module.__file__)
  mod_name = module.__name__.split('.')[-1]
  log.noise(f'resolve({module}) __file__ -> {module.__file__}')
  log.noise(f'resolve({module}) __name__ -> {module.__name__}')
  log.noise(f'resolve({module}) mod_dir -> {mod_dir}')
  log.noise(f'resolve({module}) mod_name -> {mod_name}')
  assets_dir = os.path.join(mod_dir, path.asset_dir_name)
  mod_assets_dir = os.path.join(assets_dir, mod_name)
  log.noise(f'resolve({module}) -> {mod_assets_dir}')
  return mod_assets_dir

def lookup(obj, name):
  assets_dir = path.resolve(import_module(obj.__module__))
  if not assets_dir:
    log.noise(f'lookup({obj}, {name}) __module__ -> None')
    return None
  log.noise(f'lookup({obj}, {name}) __module__ -> {obj.__module__}')
  asset = assets_dir
  if name:
    asset = os.path.join(assets_dir, name)
  if os.path.exists(asset) or os.path.islink(asset):
    log.noise(f'lookup({obj}, {name}) FOUND -> {asset}')
    return asset
  log.noise(f'lookup({obj}, {name}) MISSING -> {asset}')
  return None

def path(obj, name=None, bases=True):
  '''
  Return the absolute path for an asset. `obj` is resolved to a module, and the
  module is passed to `resolve()`. `name` is looked up within the module's
  assets directory.

  `obj` - function or class type
  `name` - file path relative to the `obj`'s module assets directory
  '''
  result = None
  if isinstance(obj, types.FunctionType):
    result = path.lookup(obj, name)
  elif inspect.isclass(obj):
    for cls in inspect.getmro(obj):
      asset = path.lookup(cls, name)
      if asset:
        result = asset
        break
  else:
    return path(type(obj), name, bases)
  if result:
    log.noise(f'path({obj}, name={name}, bases={bases}) FOUND -> {result}')
    return result
  log.noise(f'path({obj}, name={name}, bases={bases}) -> MISSING')
  raise FileNotFoundError(f"Asset '{name}' not found for obj {obj}")

path.asset_dir_name = '__assets__'
path.lookup = lookup
path.resolve = resolve

def open(obj, name, mode, encoding=None):
  log.noise(f'open({obj}, {mode}, encoding={encoding})')
  return open(path(obj, name), mode=mode, encoding=encoding)

def slurp(obj, name, mode, encoding=None):
  log.noise(f'slurp({obj}, {name}, {mode}, encoding={encoding})')
  with open(path(obj, name), mode=mode, encoding=encoding) as fd:
    return fd.read()

def loadf(obj, name, encoding=None):
  log.noise(f'loadf({obj}, {name}, encoding={encoding}')
  return formats.loadf(path(obj, name), encoding=encoding)



from lura.plates.base import Expander

class FString(Expander):
  'Expand templates as f-strings.'

  def __init__(self):
    super().__init__()

  def expands(self, env, tmpl):
    '''
    Expand a template.

    :param dict env: expansion environment
    :param str tmpl: template text
    :returns: The expanded template.
    :rtype: str
    '''
    return eval("f'''{}'''".format(tmpl.replace("'''", "\\'\\'\\'")), env)

  def expandf(self, env, tmpl, dst, encoding=None):
    '''
    Expand a template to a file.

    :param str env: expansion environment
    :param str tmpl: template text
    :param str dst: destination file path
    :returns: True if dst file was written
    '''
    with open(dst, 'w', encoding=encoding) as fd:
      fd.write(self.expands(env, tmpl))


fstring = FString()



from setuptools import setup, find_packages

name             = 'lura'
version          = '0.3.17'
author           = 'eckso'
author_email     = 'eckso@eckso.io'
url              = 'https://github.com/ecks0/lura'
description      = 'syntactic sugar'

long_description              = open('README.md').read()
long_description_content_type = 'text/markdown'

python_requires  = ">= 3.6"
install_requires = open('requirements.txt').read().strip().splitlines()

setup(
  name = name,
  version = version,
  author = author,
  author_email = author_email,
  description = description,
  long_description = long_description,
  long_description_content_type = long_description_content_type,
  python_requires = python_requires,
  install_requires = install_requires,
  packages = find_packages(),
  include_package_data = True,
)



import pickle
from lura.attrs import ottr
from lura.formats.base import Format

class Pickle(Format):

  def __init__(self):
    super().__init__()

  def loads(self, data):
    'Load pickle from string ``data``.'

    return pickle.loads(data)

  def loadf(self, src, encoding=None):
    'Load pickle from file ``src``.'

    with open(src, mode='rb', encoding=encoding) as fd:
      return self.loadfd(fd)

  def loadfd(self, fd):
    'Load pickle from file descriptor ``fd``.'

    return pickle.load(fd)

  def dumps(self, data):
    'Return dict ``data`` as pickle.'

    return pickle.dumps(data)

  def dumpf(self, data, dst, encoding=None):
    'Write dict ``data`` as pickle to file ``dst``.'

    with open(dst, mode='wb', encoding=encoding) as fd:
      self.dumpfd(data, fd)

  def dumpfd(self, data, fd):
    'Write dict ``data`` as pickle to file descriptor ``fd``.'

    pickle.dump(data, fd)
    if hasattr(fd, 'flush') and callable(fd.flush):
      fd.flush()



import inspect
import sys
import threading
from lura.hooks import Hooks
from io import StringIO

def trace(frame, ev, arg):
  if ev == 'call':
    trace.hooks.call(trace, 'call', frame, ev, arg)
  elif ev == 'exception':
    trace.hooks.exc(trace, 'exc', frame, ev, arg)

trace.hooks = Hooks()

def set_trace(enabled):
  if enabled:
    if sys.gettrace() != trace:
      sys.settrace(trace)
    if threading._trace_hook != trace:
      threading.settrace(trace)
  else:
    if sys.gettrace() == trace:
      sys.settrace(None)
    if threading._trace_hook == trace:
      threading.settrace(None)

trace.set_trace = set_trace

def get_trace():
  return sys.gettrace() == trace

trace.get_trace = get_trace

class TraceHook:

  def __init__(self, module_prefix):
    super().__init__()
    self.module_prefix = module_prefix

  def echo(self, msg):
    print(msg, file=sys.stderr)

  def get_call_data_(self, frame):
    funcname = frame.f_code.co_name
    arginfo = inspect.getargvalues(frame)
    module = None
    cls = None
    args = []
    varargs = []
    kwargs = []
    if arginfo.args:
      argnames = list(arginfo.args)
      if argnames[0] == 'self':
        _ = type(arginfo.locals['self'])
        module = _.__module__
        cls = _.__name__
        del argnames[0]
      elif argnames[0] == 'cls':
        _ = arginfo.locals['cls']
        module = _.__module__
        cls = _.__name__
        del argnames[0]
      args = [f'{_}={repr(arginfo.locals[_])}' for _ in argnames]
    if arginfo.varargs is not None:
      varargs = [repr(_) for _ in arginfo.locals[arginfo.varargs]]
    if arginfo.keywords is not None:
      kwargs = [
        f'{repr(k)}: {repr(v)}'
        for k, v in arginfo.locals[arginfo.keywords].items()
      ]
    if module is None:
      module = inspect.getmodule(frame.f_code)
      module = '__none__' if module is None else module.__name__
    return module, cls, funcname, args, varargs, kwargs

  def format_call_data_(self, module, cls, funcname, args, varargs, kwargs):
    buf = StringIO()
    w = buf.write
    if cls:
      w(f'{module}.{cls}.{funcname}(')
    else:
      w(f'{module}.{funcname}(')
    if args:
      w(', '.join(args))
    if varargs:
      if args:
        w(', ')
      w(f'*args=({", ".join(varargs)})')
    if kwargs:
      if args or varargs:
        w(', ')
      w(f'**kwargs={{{", ".join(kwargs)}}}')
    w(')')
    result = buf.getvalue()
    buf.close()
    return result

  def filter_call(self, frame, ev, arg):
    module = inspect.getmodule(frame.f_code)
    if module is None:
      return True
    modname = module.__name__
    prefix = self.module_prefix
    if isinstance(prefix, str):
      prefix = [prefix]
    return not any(modname.startswith(_) for _ in prefix)


  def call(self, o, e, frame, ev, arg):
    if self.filter_call(frame, ev, arg):
      return
    data = self.get_call_data_(frame)
    msg = self.format_call_data_(*data)
    self.echo(msg)

  def exc(self, o, e, frame, ev, arg):
    pass

trace.TraceHook = TraceHook



import logging
import logging.config
import os
import time
import types
from collections import defaultdict
from io import StringIO
from logging import NOTSET, DEBUG, INFO, WARNING, WARN, ERROR, CRITICAL, FATAL
from lura.attrs import attr
from lura.utils import DynamicProxy, asbool

class ExtraInfoFilter(logging.Filter):
  '''
  Provides additional fields to log records:

  - ``short_name`` a reasonably short name
  - ``shortest_name`` the shortest reasonably useful name
  - ``short_levelname`` a symbol associated with a level
  - ``run_time`` number of seconds the logging system has been initialized
  '''

  initialized = time.time()

  map_short_level = defaultdict(
    lambda: '=',
    DEBUG    = '+',
    INFO     = '|',
    WARNING  = '>',
    ERROR    = '*',
    CRITICAL = '!',
  )

  def filter(self, record):
    _ = record.name.split('.')
    record.short_name = '.'.join(_[-2:])
    record.shortest_name = _[-1]
    record.short_levelname = self.map_short_level.get(record.levelname)
    record.run_time = time.time() - self.initialized
    return True

class MultiLogger(DynamicProxy):

  NOTSET = logging.NOTSET
  DEBUG = logging.DEBUG
  INFO = logging.INFO
  WARNING = logging.WARNING
  WARN = logging.WARN
  ERROR = logging.ERROR
  CRITICAL = logging.CRITICAL
  FATAL = logging.FATAL

  def __init__(self, loggers):
    super().__init__(loggers)

class MultiLineFormatter(logging.Formatter):

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def format(self, record):
    if os.linesep not in record.msg:
      record.message = super().format(record)
      return record.message
    msg = record.msg
    with StringIO() as buf:
      for line in record.msg.split(os.linesep):
        record.msg = line
        buf.write(super().format(record) + os.linesep)
      record.message = buf.getvalue().rpartition(os.linesep)[0]
      #record.message = buf.getvalue().rstrip(os.linesep)
    record.msg = msg
    return record.message

class Logging:

  NOTSET = logging.NOTSET
  DEBUG = logging.DEBUG
  INFO = logging.INFO
  WARNING = logging.WARNING
  WARN = logging.WARN
  ERROR = logging.ERROR
  CRITICAL = logging.CRITICAL
  FATAL = logging.FATAL

  # Custom log formats.
  formats = attr(

    # Only the log message.
    bare = (
      '%(message)s',
      '%Y-%m-%d %H:%M:%S',
    ),

    # More old-school.
    classic = (
      '%(asctime)s %(levelname)-8s %(short_name)s %(message)s',
      '%Y-%m-%d %H:%M:%S',
    ),

    # More verbose.
    debug = (
      '%(run_time)-8.3f %(short_name)19s %(short_levelname)s %(message)s',
      '%Y-%m-%d %H:%M:%S',
    ),

    # More verbose and longer field width.
    debuglong = (
      '%(run_time)-12.3f %(name)25s %(short_levelname)s %(message)s',
      '%Y-%m-%d %H:%M:%S',
    ),

    # Run time and message.
    runtime = (
      '%(run_time)-12.3f %(message)s',
      '%Y-%m-%d %H:%M:%S',
    ),

    # Very long.
    verbose = (
      '%(asctime)s %(run_time)12.3f %(name)s %(short_levelname)s %(message)s',
      '%Y-%m-%d %H:%M:%S',
    ),
  )

  def __init__(
    self,
    std_logger,
    std_format = None,
    std_datefmt = None,
    log_envvar = None,
    level_envvar = None,
    format_envvar = None,
    append_envvar = None,
  ):
    super().__init__()
    self.std_logger = std_logger
    if std_format is None:
      std_format = self.formats.bare[0]
    if std_datefmt is None:
      std_datefmt = self.formats.bare[1]
    stdup = std_logger.upper()
    if log_envvar is None:
      log_envvar = f'{stdup}_LOG'
    if level_envvar is None:
      level_envvar = f'{stdup}_LOG_LEVEL'
    if format_envvar is None:
      format_envvar = f'{stdup}_LOG_FORMAT'
    if append_envvar is None:
      append_envvar = f'{stdup}_LOG_APPEND'
    self.std_format = std_format
    self.std_datefmt = std_datefmt
    self.log_envvar = log_envvar
    self.level_envvar = level_envvar
    self.format_envvar = format_envvar
    self.append_envvar = append_envvar
    self.config = None
    self.configure()

  def configure(self):
    if self.config is not None:
      return
    self.config_logging()
    self.config_levels()
    self.config_environment()

  def config_logging(self):
    self.config = self.build_config()
    logging.config.dictConfig(self.config)
    for level, name in logging._levelToName.items():
      setattr(logging.Logger, name, level)

  def config_levels(self):
    if 'NOISE' not in logging._levelToName.values():
      self.add_level('NOISE', 5, ':')

  def config_environment(self):
    if self.level_envvar in os.environ:
      self.set_level(os.environ[self.level_envvar].upper())
    if self.log_envvar in os.environ:
      file = os.environ[self.log_envvar]
      append = asbool(os.environ.get(self.append_envvar, '0'))
      level = os.environ.get(self.level_envvar, self.get_level())
      handler = self.add_file_handler(file, level=level, append=append)
      handler.set_name(self.log_envvar)
      fmt = os.environ.get(self.format_envvar, 'verbose').lower()
      handler.setFormatter(MultiLineFormatter(*self.formats[fmt]))

  def build_config(self):
    import yaml
    config = yaml.safe_load(f'''
      version: 1
      filters:
        short_name: {{}}
      formatters:
        standard:
          format: '{self.std_format}'
          datefmt: '{self.std_datefmt}'
      handlers:
        stderr:
          class: logging.StreamHandler
          stream: ext://sys.stderr
          filters: ['short_name']
          formatter: standard
      loggers:
        {self.std_logger}:
          handlers: ['stderr']
          level: INFO
    ''')
    config['filters']['short_name']['()'] = ExtraInfoFilter
    config['formatters']['standard']['()'] = MultiLineFormatter
    return config

  def build_logger_log_method(self, level):
    def log_level(self, msg, *args, **kwargs):
      if self.isEnabledFor(level):
        self._log(level, msg, args, **kwargs)
    return log_level

  def get_logger(self, obj=None):
    if obj is None:
      name = self.std_logger
    elif isinstance(obj, str):
      name = obj
    elif isinstance(obj, type):
      name = f'{obj.__module__}.{obj.__name__}'
    else:
      name = f'{obj.__module__}.{type(obj).__name__}'
    return logging.getLogger(name)

  getLogger = get_logger

  def get_level(self):
    'Get standard logger log level.'

    return self.get_logger(self.std_logger).getEffectiveLevel()

  def set_level(self, level):
    'Set standard logger log level.'

    if isinstance(level, str):
      level = getattr(self, level)
    self.get_logger(self.std_logger).setLevel(level)
    if level in (self.DEBUG, self.NOISE):
      self.set_formats('debug', self.std_logger)

  def get_level_name(self, number):
    return logging._levelToName[number]

  def add_level(self, name, number, short_name=None):
    if number in logging._levelToName.keys():
      msg = 'Log level number {} already used by {}'
      raise ValueError(msg.format(number, self.get_level_name(number)))
    elif name in logging._levelToName.values():
      raise ValueError(f'Log level name {name} already in use')
    logging.addLevelName(number, name)
    setattr(logging, name, number)
    setattr(logging.Logger, name, number)
    setattr(MultiLogger, name, number)
    setattr(type(self), name, number)
    setattr(
      logging.Logger, name.lower(), self.build_logger_log_method(number))
    if short_name is not None:
      ExtraInfoFilter.map_short_level[name] = short_name

  def set_format(self, format, datefmt, logger=None):
    '''
    Set the format for all handlers of a logger.

    :param str format: log format specification
    :param str datefmt: date format specification
    '''
    logger = self.std_logger if logger is None else logger
    logger = self.get_logger(logger)
    formatter = MultiLineFormatter(format, datefmt)
    for _ in logger.handlers:
      if _.get_name() == self.log_envvar:
        continue
      _.setFormatter(formatter)

  def set_formats(self, format, logger=None):
    logger = self.std_logger if logger is None else logger
    self.set_format(*self.formats[format], logger)

  def add_file_handler(self, path, level=None, logger=None, append=True):
    '''
    Add a file handler to write log messages to a file.

    :param [str, stream] path: path to log file or a stream
    :param int level: logging level, or None to defer
    :param str logger: name of logger to receive the file handler
    :param bool append: if True, append to existing log, if False, overwrite it
    :returns: the file handler, for use with ``remove_handler()``
    :rtype: logging.FileHandler
    '''
    logger = self.std_logger if logger is None else logger
    if isinstance(path, str):
      if not append and os.path.isfile(path):
        os.unlink(path)
      handler = logging.FileHandler(path)
    else:
      handler = logging.StreamHandler(path)
    if level is not None:
      if isinstance(level, str):
        level = getattr(self, level)
      handler.setLevel(level)
    formatter = MultiLineFormatter(*self.formats.verbose)
    handler.setFormatter(formatter)
    handler.addFilter(ExtraInfoFilter())
    self.get_logger(logger).addHandler(handler)
    return handler

  def remove_handler(self, handler, logger=None):
    '''
    Remove a handler from a logger.

    :param logging.Handler handler: handler instance to remove
    :param str logger: name of logger for removal
    '''
    logger = self.std_logger if logger is None else logger
    self.get_logger(logger).removeHandler(handler)

  def multilog(self, *loggers):
    'Log messages to multiple loggers.'

    return MultiLogger([_ for _ in loggers if _])



'''
Data structure helpers.

``attr`` and ``ottr`` are analogs to ``dict`` and ``OrderedDict`` which treat
attributes as keys to the backing dict.
```
>>> a = attr()
>>> a.foo = 'oof'
>>> a.foo
'oof'
>>> a['foo']
'oof'
>>> a['bar'] = 'rab'
>>> a['bar']
'rab'
>>> a.bar
'rab'
```
These classes behave like ``dict`` in every other way. Of course, only keys
which conform to Python's syntax requirements for attribute names are
accessible in this way. Traditional dict syntax must be used for keys
which cannot be used as a valid Python attribute name. Example:
```
>>> a = attr()
>>> a['foo bar'] = "python attribute names can't have spaces"
```
'''

from collections import OrderedDict, MutableMapping, defaultdict

class AttributesMixin:
  'Mixin for dict types allowing attributes to be treated as keys.'

  def __init__(self, *args, **kwargs):
    super().__init__()

  def __getattr__(self, k):
    try:
      return self[k]
    except KeyError:
      pass
    err = f"'{type(self).__name__}' object has no attribute '{k}'"
    raise AttributeError(err)

  def __setattr__(self, k, v):
    self[k] = v

class DefaultMixin:
  'Mixin for dict types implementing defaultdict behavior.'

  def __init__(self, default_factory=None, *args, **kwargs):
    super().__init__()
    self.default_factory = default_factory

  def __missing__(self, key):
    if self.default_factory is None:
      raise KeyError(key)
    value = self.default_factory()
    self[key] = value
    return value

  def __getitem__(self, k):
    try:
      return super().__getitem__(k)
    except KeyError:
      pass
    return self.__missing__(k)

  def copy(self):
    return self.__class__(self.default_factory, self)

class Attributes(dict, AttributesMixin):
  'A subclass of ``dict`` that treats attributes as keys.'

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def copy(self):
    return self.__class__(self)

  def __repr__(self):
    return f'Attributes({super().__repr__()})'

class DefaultAttributes(defaultdict, AttributesMixin):
  'A subclass of ``defaultdict`` that treats attributes as keys.'

  def __init__(self, default_factory, *args, **kwargs):
    super().__init__(default_factory, *args, **kwargs)

  def copy(self):
    return self.__class__(self.default_factory, self)

class RecursiveAttributes(DefaultAttributes):
  '''
  A subclass of ``DefaultAttributes`` that uses its own constructor as its
  factory.
  '''

  def __init__(self, *args, **kwargs):
    super().__init__(self.__class__, *args, **kwargs)

  def copy(self):
    return self.__class__(self)

class OrderedAttributes(OrderedDict, AttributesMixin):
  'A subclass of ``OrderedDict`` that treats attributes as keys.'

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

class DefaultOrderedAttributes(OrderedAttributes, DefaultMixin):
  '''
  A subclass of ``OrderedAttributes`` that implements the behavior of
  ``defaultdict``.
  '''

  def __init__(self, default_factory=None, *args, **kwargs):
    DefaultMixin.__init__(self, default_factory)
    del self['default_factory']
    OrderedDict.__setattr__(self, 'default_factory', default_factory)
    OrderedAttributes.__init__(self, *args, **kwargs)

class RecursiveOrderedAttributes(DefaultOrderedAttributes):
  '''
  A subclass of ``DefaultOrderedAttributes`` that uses its own constructor as
  its factory.
  '''

  def __init__(self, *args, **kwargs):
    super().__init__(self.__class__, *args, **kwargs)

class AttributesWrapper:

  # FIXME generalize and test; recursive AttributesWrapper

  def __init__(self, target):
    super().__init__()
    self.__dict__['__target__'] = target

  def __getitem__(self, *args, **kwargs):
    return self.__target__.__getitem__(*args, **kwargs)

  def __setitem__(self, *args, **kwargs):
    self.__target__.__setitem__(*args, **kwargs)

  def __delitem__(self, *args, **kwargs):
    self.__target__.__delitem__(*args, **kwargs)

  def __iter__(self):
    return self.__target__.__iter__()

  def __len__(self):
    return self.__target__.__len__()

  def __getattr__(self, name):
    if hasattr(self.__target__, name):
      return getattr(self.__target__, name)
    try:
      return self.__target__[name]
    except KeyError:
      pass
    err = f"'{type(self).__name__}' object has no attribute '{name}'"
    raise AttributeError(err)

  def __setattr__(self, name, value):
    if name in self.__dict__:
      self.__dict__[name] = value
      return
    target = self.__target__
    if hasattr(target, name):
      setattr(target, name, value)
    else:
      target[name] = value

attr = Attributes
dattr = DefaultAttributes
rattr = RecursiveAttributes

ottr = OrderedAttributes
dottr = DefaultOrderedAttributes
rottr = RecursiveOrderedAttributes

wttr = AttributesWrapper
